# Generated by Django 3.1.8 on 2021-05-18 08:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_twined", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="abstractregisteredservice",
            name="slug",
            field=models.SlugField(
                editable=False,
                help_text="Service name (e.g. in cloud run), also used in django settings to define service credentials",
                unique=True,
            ),
        ),
    ]



# Generated by Django 3.1.8 on 2021-05-10 18:53

import uuid
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="AbstractRegisteredService",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "slug",
                    models.SlugField(
                        help_text="Service name slug used in django settings to define service credentials"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="RegisteredService",
            fields=[
                (
                    "abstractregisteredservice_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_twined.abstractregisteredservice",
                    ),
                ),
            ],
            bases=("django_twined.abstractregisteredservice",),
        ),
        migrations.CreateModel(
            name="Question",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "registered_service",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="questions",
                        to="django_twined.registeredservice",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



import json
import logging
import uuid
from django.db import models
from octue.resources import Datafile
from octue.utils.encoders import OctueJSONEncoder

from .querysets import DatastoreQueryset


logger = logging.getLogger(__name__)


class AbstractSynchronisedDatastore(models.Model):
    """Rows in the database correspond to files in an object store, with metadata synchronized for search/querying

    The most obvious question is:
        *"Why don't I just keep the metadata in the DB table and the object in the store, like normal?"*

    Some good answers:
        - because that separates the file objects and their metadata
        - because interaction with the database is then required by every service that wants to use the data, which
            - limits scalability of the system
            - requires the dba / backend team to weigh in heavily on every application that uses or creates data
            - precludes creation of data by untrusted or third party services
            - requires the dba / backend team to design and/or have deep knowledge of the intrinsic data type contained
              in the files in order to design appropriate and consistent metadata for them

    Yes, sync to a database like this is a duplication of the metadata and that requires an additional job to maintain
    consistency, which is a pain to maintain that one synchronisation system, but not an unscalable activity.

    You can think of this like you'd think about search indexes. In order to make data items searchable you'd include
    them in a search index, which is a duplicate of metadata. Here, we're duplicating metadata in order to make our data
    items queryable, relatable to other entities AND searchable (with postgres full text search, or solutions like
    ZomboDB or django-haystack to integrate elasticsearch).

    Of course you can also use your own business logic/auth and permission systems like django guardian to control
    access and visibility of your data items however you want.

    """

    __TAG_FIELDS__ = set()
    __FILE_FIELD__ = "file"

    id = models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True)

    objects = DatastoreQueryset.as_manager()

    class Meta:
        abstract = True

    @property
    def project_name(self):
        return self._storage.project_id

    @property
    def datafile(self):
        """The octue datafile associated with this instance"""

        if self._state.adding:
            # TODO Once https://github.com/octue/octue-sdk-python/issues/200 is resolved, we will be able to
            # Retun datafiles from the cloud prior to calling () save on a new instantiated model.
            return None

        # Handle event of instance has unset file or file not yet in store
        if self._location is None:
            raise ValueError("Instance has no file field set, cannot create Datafile")

        return Datafile(self.gs_path, project_name=self._storage.project_id)

    @classmethod
    def from_datafile(cls, datafile, create_if_missing=True, update_db_metadata=True):
        """Gets database record corresponding to an octue datafile object.

        The returned instance is always updated with metadata from the datafile object, however, this updated
        instance will only be saved to the database if update_db_metadata is true, so it is possible to return a dirty
        instance from this method.

        :param datafile: Datafile object representing a cloud or local file and its metadata
        :type datafile: octue.Datafile
        :param create_if_missing: Create a database entry for this file and its metadata if it's not present
        :type create_if_missing: bool
        :param update_db_metadata: If database entry for this file is found, ensure the metadata in that row is up to date with metadata from the cloud store
        :type update_db_metadata: bool
        :return: Instance of db model
        :rtype: AbstractSynchronisedDatastore
        """

        created = False
        try:
            instance = cls.objects.get(id=datafile.id)
            if update_db_metadata:
                instance.update_instance_from_tags(datafile.tags)
                instance.update_instance_from_labels(datafile.labels)
                instance.save()

        except cls.DoesNotExist:
            instance = cls(id=datafile.id)
            instance.update_instance_from_tags(datafile.tags)
            instance.update_instance_from_labels(datafile.labels)

            # TODO see https://github.com/octue/octue-sdk-python/issues/204 - once closed, a better way of doing this will be accessible
            if datafile.path.startswith("gs://"):
                path_in_bucket = datafile.path.replace("gs://", "").split("/", 1)[1]
            else:
                path_in_bucket = datafile.path.lstrip("/")

            # Set the file field .name attribute directly to path (https://stackoverflow.com/a/10906037/3556110)
            getattr(instance, cls.__FILE_FIELD__).name = path_in_bucket

            if create_if_missing:
                instance.save()
                created = True

        return instance, created

    def get_tags_from_instance(self):
        """Get tags (a dict of key-value pairs) from this instance

        Override this method to customise how Datafile tags are created from model instance fields (or other data)
        """
        tags = dict((name, getattr(self, name)) for name in self.__TAG_FIELDS__)

        # TODO A ***FAR*** LESS TERRIBLE SERIALIZER THAN THIS BASED ON AVRO OR SIMILAR!!!
        for k, v in tags.items():
            if v.__class__.__name__ == "Decimal":
                tags[k] = float(v)

        return tags

    def get_labels_from_instance(self):
        """Get labels as a set of strings from this instance

        Override this method to customise how Datafile labels are created from the instance fields (or other data)
        """
        return set()

    def update_instance_from_tags(self, tags):
        """Set this instance's fields using provided tags (a dict or TagDict of key-value pairs)

        Override this method to customise how metadata gets updated to the model instance (from the Datafile)
        """
        for name in self.__TAG_FIELDS__:
            setattr(self, name, tags.get(name, None))

    def update_instance_from_labels(self, labels=None):
        """Update this instance's fields using provided labels (a set of strings)

        Override this method to customise how metadata gets updated to the model instance (from the Datafile)
        """
        pass

    @property
    def _file_field(self):
        return getattr(self, self.__FILE_FIELD__)

    @property
    def _storage(self):
        return self.get_storage()

    @property
    def _storage_settings(self):
        return self.get_storage_settings()

    @classmethod
    def get_storage(cls):
        mapped_fields = dict((field.attname, field) for field in cls._meta.concrete_fields)
        return mapped_fields[cls.__FILE_FIELD__].storage

    @classmethod
    def get_storage_settings(cls):
        mapped_fields = dict((field.attname, field) for field in cls._meta.concrete_fields)
        return mapped_fields[cls.__FILE_FIELD__].store["storage_settings"]

    @property
    def _location(self):
        if self._storage.location != "":
            return self._storage.location
        elif self._file_field.name != "":
            return self._file_field.name
        else:
            return None

    @property
    def gs_path(self):
        return f"gs://{self._storage.bucket_name}/{self._location}"

    def _todo_expose_on_sdk(self, metadata):
        """Encode metadata as a dictionary of JSON strings."""
        if not isinstance(metadata, dict):
            raise TypeError(f"Metadata for Google Cloud storage should be a dictionary; received {metadata!r}")

        return {key: json.dumps(value, cls=OctueJSONEncoder) for key, value in metadata.items()}

    def save(self, *args, **kwargs):
        """Override the save method to synchronize metadata to the cloud store"""

        # Create a hypothetical (doesn't care if it's remote or local, or never existing)
        #  datafile instance simply to generate octue-compliant metadata
        datafile = Datafile(
            self._location,
            id=self.id,
            tags=self.get_tags_from_instance(),
            labels=self.get_labels_from_instance(),
            hypothetical=True,
        )
        metadata = self._todo_expose_on_sdk(datafile.metadata())

        # Only commit to saving if saving the metadata works correctly too
        self.file.file.metadata = metadata
        super().save(*args, **kwargs)

        # Update the metadata that we added (id, tags, labels) above
        datafile = self.datafile
        datafile.tags = self.get_tags_from_instance()
        datafile.labels = self.get_labels_from_instance()
        datafile.update_cloud_metadata()



import logging
from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand


logger = logging.getLogger(__name__)


VERBOSITY_MAP = {
    0: logging.ERROR,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}


class Command(BaseCommand):
    """Use `python manage.py help sync_data_stores` to display help for this command line administration tool"""

    help = (
        "Syncs files and their metadata from stores (defined in SETTINGS) into queryable records in the database."
        " Use in development for repopulating an empty database with the contents of the store(s)."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--source-keys",
            nargs="+",
            required=False,
            dest="source_keys",
            default=None,
            help="Specify the data store keys. If none, all keys specified in settings.TWINED_DATA_STORES will be used.",
        )

    def handle(self, *args, source_keys=None, **options):

        # Ensure we respect the --verbosity command option
        verbosity = int(options["verbosity"])
        logger.setLevel(VERBOSITY_MAP[verbosity])

        # Get the data sources to synchronise (by default, all defined)
        source_keys = source_keys or settings.TWINED_DATA_STORES.keys()
        stores = dict((key, settings.TWINED_DATA_STORES[key]) for key in source_keys)

        # Loop through the data sources
        for key, store in stores.items():
            logger.info(f"Synchronizing store -> database for {key}")

            # Get model class and the storage instance used for its datafiles
            Model = apps.get_model(*store["model"].split("."))

            # Daisychain sync commands so the store comparison only gets run once
            Model.objects.import_missing().delete_missing().sync_metadata_from_store()



import django.dispatch


question_sent = django.dispatch.Signal()

answer_received = django.dispatch.Signal()

monitor_updated = django.dispatch.Signal()

logstream_updated = django.dispatch.Signal()


class ServiceSignaler:
    def send_question_sent(self, question):
        question_sent.send(sender=self.__class__, question=question)

    def send_answer_received(self, answer):
        answer_received.send(sender=self.__class__, answer=answer)

    def send_monitor_updated(self, monitor):
        monitor_updated.send(sender=self.__class__, monitor=monitor)

    def send_logstream_updated(self, logentry):
        logstream_updated.send(sender=self.__class__, logentry=logentry)



import logging
from django.conf import settings
from django.db import models
from django.utils.module_loading import import_string


logger = logging.getLogger(__name__)


class DatafileObjectField(models.FileField):
    """A FileField allowing storage class to be declared in settings

    This enables storages to be switched without database migrations, and enables different FileFields to be associated
    with different storages rather than them all using the django media bucket.
    """

    def __init__(self, store_key=None, **kwargs):
        if store_key is None:
            raise ValueError(
                "Missing store_key argument. Your Datastore model must contain a `file = DatafileObjectField(store_key='...')` field declaration"
            )

        self.store_key = store_key
        store = settings.TWINED_DATA_STORES[store_key]
        kwargs["storage"] = import_string(store["storage"])(**store["storage_settings"])
        super().__init__(**kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["store_key"] = self.store_key
        kwargs.pop("storage")
        return name, path, args, kwargs


class DatafileMetadataField(models.JSONField):
    """Contains metadata not otherwise present in"""

    def __init__(self, *args, **kwargs):
        kwargs["default"] = kwargs.pop("default", dict)
        kwargs["help_text"] = kwargs.pop("help_text", "Additional object metadata")
        super().__init__(*args, **kwargs)


class ValuesField(models.JSONField):
    """Values storage in the actual DB
    TODO accept schema for validation and render better forms
    """

    def __init__(self, *args, **kwargs):
        kwargs["default"] = kwargs.pop("default", dict)
        kwargs["help_text"] = kwargs.pop("help_text", "Values data")
        # By default, work around https://code.djangoproject.com/ticket/27697
        kwargs["blank"] = kwargs.pop("blank", True)
        super().__init__(*args, **kwargs)


class ManifestField(models.JSONField):
    """Manifest storage in the actual DB
    TODO accept schema for validation and render better forms
    """

    def __init__(self, *args, **kwargs):
        kwargs["default"] = kwargs.pop("default", dict)
        kwargs["help_text"] = kwargs.pop("help_text", "Manifest data")
        super().__init__(*args, **kwargs)



import logging
from uuid import uuid4
from asgiref.sync import async_to_sync
from channels.generic.websocket import WebsocketConsumer
from django.core.cache import cache
from django_twined.messages import ReelMessage
from django_twined.tasks import ask


logger = logging.getLogger(__name__)


class AnalysisConsumer(WebsocketConsumer):
    """A consumer allowing creation and monitoring of analyses


    ~~ When in danger or in doubt, run in circles, scream and shout ~~


    I want to: load a configuration at startup of the server.
    I want to: load a frontend which knows about my twine and configuration settings.
    I want to: connect my frontend to a backend.
    I want to: supply a backend with input data periodically.
            - I can write a quick python script to do this from local.
    I want to: use the frontend to show some graphic of the most recent input data, continuously updating.
    I want to: use the frontend to show an input form.
    I want to: use the frontend to show an output (disabled) form with results.

    """

    def connect(self):
        """Accept connection to the analysis, and forward any prior messages to the client"""
        self.group_id = str(self.scope["url_route"]["kwargs"].get("analysis_id", uuid4()))

        # Set the group name
        self.group_name = f"analysis-{self.group_id}"

        # Add this channel to the group
        async_to_sync(self.channel_layer.group_add)(self.group_name, self.channel_name)

        # Accept the connection
        self.accept()

        # Flush the cache of messages from this group so far
        for message in cache.get(self.group_name, default=list()):
            self.send(text_data=message)

    def disconnect(self, close_code):
        """Accept a disconnection from the analysis gracefully"""
        async_to_sync(self.channel_layer.group_discard)(self.group_name, self.channel_name)

    def receive(self, text_data=None, bytes_data=None):
        """Receive service instructions from WebSocket"""

        # Get the incoming message data
        try:
            message = ReelMessage(src=text_data)
        except Exception as e:
            logger.warning(f"Possible error (or malformed message): {e.args[0]}")
            ReelMessage(status="error", hints="Unable to understand your message").send(self)
            return

        # Attempt to respond
        try:
            # Run a new analysis
            if message.action == "ask":
                ask(self.group_id, message).send(self)

        except Exception as e:
            logger.error(e.args[0])
            ReelMessage(
                action=message.action,
                status="error",
            ).send(self)
            return

        # # Append changes to the cache for this room, allowing us to flush all changes through to new connection
        # # TODO Race condition exists here. Either rewrite automerge for python and keep a master record, or use
        # #  a locking strategy enabling us to ensure all changes are recorded or use a synchronous websocketconsumer
        # #  to properly persist the change history
        # #  https://pypi.org/project/python-redis-lock/
        # #  https://github.com/joanvila/aioredlock
        # changes = cache.get(self.room_name, default=list())
        # changes.append(text_data)
        # cache.set(self.room_name, changes)
        # data = {
        #     "type": "reel_message",
        #     "message": text_data,
        #     "message_bytes": bytes_data,
        # }
        # await self.channel_layer.group_send(self.room_group_name, data)

    def reel_message(self, event):
        """Send a message from the analysis to the client, serialising if necessary"""
        self.send(text_data=event["message"])



# Generated by Django 3.1.8 on 2021-05-18 09:23

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("django_twined", "0002_slug_unique_and_not_editable"),
    ]

    operations = [
        migrations.RenameField(
            model_name="abstractregisteredservice",
            old_name="slug",
            new_name="name",
        ),
    ]



from django.contrib import admin

from .models import RegisteredService


class RegisteredServiceAdmin(admin.ModelAdmin):
    search_fields = ["id", "name"]
    list_display = ("id", "name")
    readonly_fields = ("id", "name")


admin.site.register(RegisteredService, RegisteredServiceAdmin)



import logging
import uuid
from django.apps import apps
from django.conf import settings
from django.db import models
from octue.cloud.pub_sub.service import Service
from octue.resources.service_backends import get_backend


logger = logging.getLogger(__name__)


class AbstractRegisteredService(models.Model):
    """Abstract model to enable registration of available services in the system"""

    __QUESTION_MODEL__ = "django_twined.Question"

    id = models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True)

    name = models.SlugField(
        editable=False,
        help_text="Service name (e.g. in cloud run), also used in django settings to define service credentials",
        unique=True,
    )

    # TODO open an endpoint to receive updates of configuration_values, configuration_manifest and twine JSON
    # configuration_values = models.JSONField(default=dict(), help_text="Contents of the configuration_values strand")
    # configuration_manifest = models.JSONField(default=dict(), help_text="Contents of the configuration_manifest strand")

    @classmethod
    def get_question_model(cls):
        """Returns the question model used to store results"""
        return apps.get_model(*cls.__QUESTION_MODEL__.split("."))

    @property
    def topic(self):
        return f"octue.services.{self.id}"

    @property
    def service_settings(self):
        return settings.OCTUE_SERVICES[self.name]

    def ask(self, wait=False, timeout=1000, save_question=True, **kwargs):
        """Ask this service a question
        ```
        RegisteredService.objects.get(name='my-service').ask(input_values={}, input_manifest={})
        ```
        """
        backend = get_backend(self.service_settings["backend"])(
            project_name=self.service_settings["project_id"],
            credentials_environment_variable=self.service_settings["credentials_environment_variable"],
        )

        asker = Service(backend)
        subscription, question_uuid = asker.ask(
            service_id=self.topic,
            **kwargs,
        )

        QuestionModel = self.get_question_model()
        question = QuestionModel(id=question_uuid, registered_service=self, **kwargs)

        if save_question:
            question.save()

        if wait:
            question.update_answer(asker.wait_for_answer(subscription, timeout))
            if save_question:
                question.save()

        return question, subscription

    def __str__(self):
        return f"{str(self.id)[:12]} {self.name}"


class RegisteredService(AbstractRegisteredService):
    pass



# Disable these to allow import of the signals on app eady callback
# pylint: disable=import-outside-toplevel,unused-import

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DjangoTwinedAppConfig(AppConfig):
    """Configuration and signal registration for the django_twined app"""

    name = "django_twined"
    label = "django_twined"
    verbose_name = _("Django Twined")

    def ready(self):
        import django_twined.signals  # noqa:F401



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_twined',
 'django_twined.consumers',
 'django_twined.management',
 'django_twined.management.commands',
 'django_twined.migrations',
 'django_twined.models',
 'django_twined.models.querysets',
 'django_twined.signals']

package_data = \
{'': ['*'], 'django_twined': ['static/*', 'templates/django_twined/*']}

install_requires = \
['channels>=3.0,<4.0',
 'django>=3.0,<4.0',
 'octue==0.10.4',
 'pika>=1.2.0,<2.0.0']

setup_kwargs = {
    'name': 'django-twined',
    'version': '0.2.1',
    'description': 'A django app to manage octue services',
    'long_description': "[![PyPI version](https://badge.fury.io/py/django-twined.svg)](https://badge.fury.io/py/django-twined)\n[![codecov](https://codecov.io/gh/octue/django-twined/branch/master/graph/badge.svg)](https://codecov.io/gh/octue/django-twined)\n[![Documentation Status](https://readthedocs.org/projects/django-twined/badge/?version=latest)](https://django-twined.readthedocs.io/en/latest/?badge=latest)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n[![black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n\n# Django Twined\n\nThis is a plugin for django, enabling you to orchestrate [twined-based octue services](https://octue.readthedocs.io) services from your own django\nserver.\n\nDocumentation is [here](https://django-twined.readthedocs.io/en/latest/).\n\nThis is great for advanced use cases where:\n\n- you have specific security/firewalling requirements, or\n- you want to manage your own auth, or\n- you have specific/unusual data integration needs, or\n- already have a web based data service, and want to expose it in the twined ecosystem.\n\n**Health warning:** to use this plugin to deploy your services, you'll need to handle all your own data\nstorage/orchestration, devops, server management, security and auth. So for most users we'd recommend getting in touch at\n[octue.com](https://www.octue.com/contact) so we can help!\n",
    'author': 'Tom Clark',
    'author_email': 'tom@octue.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/octue/django-twined',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from django_twined.models import RegisteredService


logger = logging.getLogger(__name__)
User = get_user_model()


class Command(BaseCommand):
    """Use `python manage.py help register_services` to display help for this command line administration tool"""

    help = "Registers the services available in settings into the database, enabling relation to services and asks."

    def handle(self, *args, **options):

        for service in settings.TWINED_SERVICES:
            RegisteredService.objects.get_or_create(id=service["topic_id"], name=service["service_name"])



import json
import logging
from time import time
from uuid import uuid4
import django.dispatch
from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer


logger = logging.getLogger(__name__)

reel_message_sent = django.dispatch.Signal()

channel_layer = get_channel_layer()


MESSAGE_FIELDS = (
    "action",
    "children",
    "configuration_manifest",
    "configuration_values",
    "credentials",
    "data",
    "hints",
    "input_manifest",
    "input_values",
    "logs",
    "output_manifest",
    "output_values",
    "reference",
    "status",
    "timestamp",
    "twine",
    "uuid",
    "value",
)


class ReelMessage:
    """Class to manage incoming and outgoing messages"""

    def __init__(self, src=None, **kwargs):
        """Constructor for ServiceMessage"""
        if src is not None:
            kwargs = {
                **json.loads(src),
                **kwargs,
            }
            # TODO schema based validation with no extra fields allowed

        if "action" not in kwargs.keys():
            raise Exception("No action specified")

        kwargs["status"] = kwargs.pop("status", "success")

        # Check no uuid supplied and create one for this message
        if kwargs.get("uuid", None) is not None:
            raise ValueError("Cannot instantiate a reel message with a UUID. Each message must have a unique UUID.")
        kwargs["uuid"] = str(uuid4())

        # Add a server timestamp
        if kwargs.get("timestamp", None) is not None:
            raise ValueError(
                "Cannot instantiate a reel message with a timestamp. Each message must determine its timestamp."
            )
        kwargs["timestamp"] = time()

        self.__dict__ = dict((k, kwargs.get(k, None)) for k in MESSAGE_FIELDS)

    def serialise(self):
        """Serialise self to a string"""
        to_serialise = dict((k, getattr(self, k)) for k in MESSAGE_FIELDS if getattr(self, k) is not None)
        return json.dumps(to_serialise)

    def send(self, obj):
        """Send this message to an individual channel name
        Also sends a signal allowing 3rd party apps to execute callbacks for certain messages
        """
        serialised = self.serialise()
        logger.debug("Sending ReelMessage to channel '%s': %s", obj, serialised)
        obj.send(serialised)
        reel_message_sent.send(sender=self.__class__, message=self, to_channel=obj, to_group=None)

    def group_send(self, group_name, message_type="reel_message"):
        """Send this message to a group over channels
        Also sends a signal allowing 3rd party apps to execute callbacks for certain messages
        """
        serialised = self.serialise()
        logger.debug("Group sending ReelMessage: %s", serialised)
        async_to_sync(channel_layer.group_send)(group_name, {"type": message_type, "message": serialised})
        reel_message_sent.send(sender=self.__class__, message=self, to_channel=None, to_group=group_name)



import logging
import uuid
from django.db import models

from ..fields import ManifestField, ValuesField


logger = logging.getLogger(__name__)


class AbstractQuestion(models.Model):

    id = models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True)
    asked = models.DateTimeField(null=True, blank=True, help_text="When the question was asked")
    answered = models.DateTimeField(null=True, blank=True, help_text="When the question was answered")

    class Meta:
        abstract = True


class Question(AbstractQuestion):
    registered_service = models.ForeignKey(
        "django_twined.RegisteredService",
        related_name="questions",
        on_delete=models.PROTECT,
    )


class QuestionValuesDatabaseStorageMixin:
    """Use this mixin to store question-and-answer values data in the actual database"""

    input_values = ValuesField(help_text="Contents of the input_values strand")

    output_values = ValuesField(help_text="Contents of the output_values strand")


class QuestionManifestsDatabaseStorageMixin:
    """Use this mixin to store question-and-answer manifest data in the actual database"""

    input_manifest = ManifestField(help_text="Contents of the input_manifest strand")

    output_manifest = ManifestField(help_text="Contents of the output_manifest strand")



from .datastores import AbstractSynchronisedDatastore
from .questions import AbstractQuestion, Question
from .services import AbstractRegisteredService, RegisteredService


__all__ = (
    "AbstractRegisteredService",
    "AbstractSynchronisedDatastore",
    "AbstractQuestion",
    "RegisteredService",
    "Question",
)
